/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <math.h>
#include "ftypes.h"
#include "prec.h"
#include "exp4.h"
#include "errors.h"
#include "debug.h"

#include "hbounds.h"


#if (defined(SINGLE_PRECISION) || defined(SINGLE_COMPLEX))
static Real Tend = FLT_MAX;
static Real TForce = FLT_MAX;
static Real Hmax = FLT_MAX;
#else
static Real Tend = DBL_MAX;
static Real TForce = DBL_MAX;
static Real Hmax = DBL_MAX;
#endif

static int UpperBounded = 0;

void PREC(Exp4SetHmax) (Real HmaxIn)
{
  Hmax = fabs(HmaxIn);
}

void PREC(Exp4SetTend) (Real TendIn)
{
  Tend = TendIn;
}

void PREC(Exp4ForceTime) (Real TForceIn)
{
  TForce = TForceIn;
}

Real PREC(Exp4ChkHBounds) (Real h,Real ActualTime)
{
  if (h >= 0.0)
  {
    h = min(Hmax,h);

    if (ActualTime + h > Tend)
      h = Tend - ActualTime;

    if ((ActualTime + h > TForce) && (ActualTime < TForce))
      h = TForce - ActualTime, UpperBounded = 1;
  }
  else
  {
    h = max(-Hmax,h);

    if (ActualTime + h < Tend)
      h = Tend - ActualTime;

    if ((ActualTime + h < TForce) && (ActualTime > TForce))
      h = TForce - ActualTime, UpperBounded = 1;
  }

  return h;
}

int PREC(Exp4ChkHIsUpperBounded)()
{
  int tmp = UpperBounded;
  UpperBounded = 0;
  return tmp;
}
