/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "ftypes.h"
#include "prec.h"
#include "exp4.h"
#include "errors.h"
#include "krylov.h"
#include "history.h"

/* #define LOCAL_DEBUG_LEVEL 40 */
#define LOCAL_DEBUG_OFFSET 20
#include "debug.h"

#include "krystep.h"

Real PREC(GetKryFac) (krylov *Kry,History *Hist)
{
  finteger KryDim0,KryDim1,KryDim2,KryDim3;
  finteger help;
  Real Fac= 1.0;

  KryDim0 = KrylovSize(Kry);
  KryDim1 = GetHistoryEntry(Hist,0);
  KryDim2 = GetHistoryEntry(Hist,-1);
  KryDim3 = GetHistoryEntry(Hist,-2);

  help = max(KryDim0,KryDim1);
  if (help < KrylovGoodStep(Kry)) {
    (Fac) = pow(KrylovOptStep(Kry)/((fdouble)(max(help,1))),
	  KrylovExponent(Kry));

    help = max(KryDim1,KryDim2);
    if (max(KryDim0,help) < KrylovLowStep(Kry)) {
      (Fac) = 4.0;
      if (KryDim3 < KrylovLowStep(Kry))
	(Fac) = 8.0;
    }
  }

  Exp4Message2((DEBUG_FILE,"Krystep: History: %3d %3d %3d [%3d] => Fac = %e\n",
	KryDim3,KryDim2,KryDim1,KryDim0,Fac));

  return Fac;
}

int PREC(GetNearestIndex) (fdouble Val,finteger *index,int indexlen)
{
  int i;
  double diff,lastdiff;

  lastdiff = Val;
  for (i=0; i < indexlen; i++) {
    diff = fabs(Val - index[i]);
    if (lastdiff < diff)
      return index[i-1];
    lastdiff = diff;
  }

  return (Val/2);
}
