/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include "ftypes.h"
#include "fblas.h"
#include "exp4.h"
#include "prec.h"
#include "work.h"

#include "phi_ex.h"

static finteger inc1 = 1;

static Complex one       = ComplexConst(1.0);
static Complex zero      = ComplexConst(0.0);
static RealCoef one_three = RealCoefConst(1.0/3.0);
static RealCoef two_three = RealCoefConst(2.0/3.0);

void PREC(phi_ex) (WorkSpace Work,Complex *k__,finteger *ldk,
      int size,Complex *fy,finteger *neq,Complex *A__,
      finteger *ldA,Complex *phiA__,finteger *ldphiA,Real h)
{
#define k(I)     (k__+((I)-1)*(*ldk))
#define A(I,J)     A__[(I) + (J)*(*ldA)]
#define phiA(I,J)  phiA__[(I) +(J)*(*ldphiA)]

  Complex help;
  Complex *temp;

  DistributeWorkspace(temp,sizeof(Complex)*(*neq),Work);

  /* k(1) = phiA*fy;*/
  pgemv ("No transpose",neq,neq,&one,&phiA(0,0),ldphiA,fy,&inc1,
	&zero,k(1),&inc1);

  if (size > 1)
  {
    /*k(2) = (h/6)*(phiA*(A*k(1))) + k(1);*/
    /* copy k(1) to k(2) */
    pcopy (neq,k(1),&inc1,k(2),&inc1);

    /* put h/6*A*k(1) into temp */
    ComplexSetReal(help,h/6.0);
    pgemv ("No transpose",neq,neq,&help,&A(0,0),ldA,k(1),&inc1,
	  &zero,temp,&inc1);

    /* add h/6*phiA*(A*k(1))=phiA*work to k(2) */
    pgemv ("No transpose",neq,neq,&one,&phiA(0,0),ldphiA,temp,&inc1,
	  &one,k(2),&inc1);


    /* k(3) = (2*h/9)*( phiA*(A*k(2)) ) + (2/3)*k(2) + (1/3)*k(1); */
    /* copying k(1) to k(3) */
    pcopy (neq,k(1),&inc1,k(3),&inc1);

    /* scaling k(3) with 1/3 */
    prscal (neq,&one_three,k(3),&inc1);

    /* add 2/3*k(2) to k(3) */
    praxpy (neq,&two_three,k(2),&inc1,k(3),&inc1);

    /* put h*2/9*A*k(2) into temp */
    ComplexSetReal(help,h*2.0/9.0);
    pgemv ("No transpose",neq,neq,&help,&A(0,0),ldA,k(2),&inc1,
	  &zero,temp,&inc1);

    /* add h*2/9*phiA*work=phiA*(A*k(2)) to k(3) */
    pgemv ("No transpose",neq,neq,&one,&phiA(0,0),ldphiA,temp,&inc1,
	  &one,k(3),&inc1);
  }
#undef A
#undef phiA
#undef k
}
