/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

#include <limits.h>
#include <math.h>
#include "ftypes.h"
#include "fblas.h"
#include "flapack.h"
#include "exp4.h"
#include "prec.h"
#include "work.h"

/* #define LOCAL_DEBUG_LEVEL 100 */
#define LOCAL_DEBUG_OFFSET 10
#include "debug.h"

#include "phi.h"
#include "phimhs.h"
#include "phimge.h"
#include "phie1.h"

#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
#define PORMQR_TRANSPOSE "Transpose"
#else
#define PORMQR_TRANSPOSE "Conjugate transpose"
#endif

#define USE_QR_DECOMP_IN_DIAG
/* #define USE_HESSENBERG_FORM */

#define PADE_N       6

/* static void bandedmm(finteger roco,Complex *res,finteger ldres, */
/*       Complex *a,finteger lda,finteger bandau,finteger bandal, */
/*       Complex *b,finteger ldb,finteger bandbu,finteger bandbl); */

/* the pade coeff's */
/* static RealCoef pade_n[PADE_N]={ */
/*   RealCoefConst(1.0/26.0),RealCoefConst(5.0/156.0), */
/*   RealCoefConst(1.0/858.0),RealCoefConst(1.0/5720.0), */
/*   RealCoefConst(1.0/205920.0),RealCoefConst(1.0/8648640.0)}; */
/* static RealCoef pade_d[PADE_N]={ */
/*   RealCoefConst(-6.0/13.0),RealCoefConst(5.0/52.0), */
/*   RealCoefConst(-5.0/429.0),RealCoefConst(1.0/1144.0), */
/*   RealCoefConst(-1.0/25740.0),RealCoefConst(1.0/1235520.0)}; */

#define MAX_COND     1E5

#define MAX_FULL     11
#define MAX_BAND     3

/* only short for INT_MAX, done because the table below is more pretty */
#define IMX  INT_MAX

static int max_dim_he[MAX_BAND+1][MAX_FULL+1] = {
/* s = 0   1   2   3   4   5   6   7   8   9   10 >11 */
     {IMX,IMX,IMX,IMX,IMX,100, 50, 30, 20, 17, 11,  0},  /* bandwith = 1 */
     {IMX,IMX,IMX,IMX,200,100, 46, 28, 20, 15, 11,  0},  /* bandwith = 2 */
     {IMX,IMX,500,300,100, 80, 40, 25, 18, 13, 10,  0},  /* bandwith = 3 */
     {IMX,200,100, 60, 46, 23, 17, 13, 11,  8,  7,  0}}; /* bandwith > 3 */

static int max_dim_ge[12]    =
     {IMX,100,60, 46,23,19,17,14,11,9,7 , 0};

static finteger inc0 = 0;
static Complex one  = ComplexConst(1.0);
static Complex zero = ComplexConst(0.0);
static RealCoef one_three = RealCoefConst(1.0/3.0);
static Real two_three  = 2.0/3.0;
static finteger inc1=1;

size_t PREC(PhiZSizeWorkFun) (finteger MaxDim)
{
  /* FIXME verify Workspace calculation */
  size_t SelfSize = GetWorkSpaceSize(sizeof(Complex)*MaxDim*MaxDim);
  size_t ShareSize = GetWorkSpaceSize(sizeof(Complex)*MaxDim);
  /* FIXME fac_t is here hardcoded equal three */
  size_t Phie1HsDiagSize = PREC(SizePhie1HsDiagWork) (MaxDim,3);
  size_t PadeGeSize = PREC(SizePhimGePadeWork) (min(MaxDim,MAX_FULL));
  size_t PadeHsSize = PREC(SizePhimHsPadeWork) (MaxDim);

  size_t PhimSize = max(max(Phie1HsDiagSize,PadeGeSize),PadeHsSize);

  return SelfSize + max(ShareSize,PhimSize);
}

void PREC(Exp4PhiZ) (WorkSpace Work,Real alpha,
      finteger nZColumns,Complex *Z_,finteger ldZ,Real beta,
      finteger k,finteger HBandWith,Complex *H,finteger ldH)
{
#define Z(i) (Z_+(i)*ldZ)
  if (k == 1)
  {
    finteger i;
    for (i=1;i<=nZColumns;i++)
    {
      Complex z = *H;
      rmul (z,beta*i,z);
      z = PREC(phi) (z);
      rmul(*Z(i-1),alpha,z);
    }
  }
  else
  {
    Complex *PhiH;
    Complex *temp;
    finteger ldPhiH = k;
    int result = 0;
    int s;
    Real normH;
    finteger ldH1 = ldH+1;
    finteger bandwith = min(HBandWith,k-1);
    
    { /* calculates the inf-norm of A */
      Real *NrmWork;
      DistributeWorkspace(NrmWork,sizeof(Real)*k,Work);
      normH = plangb("Inf",&k,&inc1,&bandwith,H-bandwith,&ldH1,NrmWork);
      GiveBackWorkspace(NrmWork,sizeof(Real)*k,Work);
    }
    Exp4Message2((DEBUG_FILE,"|| H || = %e\n",normH));
    
    /* get the scaling s */
    frexp(normH*beta,&s);
    s = max(0,s+1);
    
    if (k <= MAX_FULL)
    {
      /* If s is too big take the diagonalization routine.
	 If the transformation to diagonal form is ill conditioned,
	 result is equal 0, then use the pade approximation anyway. */
      if (k > max_dim_ge[min(s,MAX_FULL)])
	result = PREC(Exp4Phie1HsDiag) (Work,Z_,ldZ,nZColumns,H,ldH,k,
	      beta,alpha);
      if (result == 0)
      {
	DistributeWorkspace(PhiH,sizeof(Complex)*k*ldPhiH,Work);
	result = PREC(Exp4PhimGePade)(Work,PhiH,ldPhiH,H,ldH,k,s,beta);
      }
      else
	return;
    }
    else
    {
      /* If s is too big take the diagonalization routine.
	 If the transformation to diagonal form is ill conditioned,
	 result is equal 0, then use the pade approximation anyway. */
      if (k > max_dim_he[max(min(bandwith,MAX_BAND)-1,0)][min(s,MAX_FULL)])
	result = PREC(Exp4Phie1HsDiag) (Work,Z_,ldZ,nZColumns,H,ldH,k,
	      beta,alpha);
      if (result == 0)
      {
	DistributeWorkspace(PhiH,sizeof(Complex)*k*ldPhiH,Work);
	result = PREC(Exp4PhimHsPade)(Work,PhiH,ldPhiH,H,ldH,k,
	      bandwith,s,beta);
      }
      else
	return;
    }

    DistributeWorkspace(temp,sizeof(Complex)*k,Work);
    
    /* z(1:k,1) = alpha*phiH(:,1) */
    pcopy (&k,PhiH,&inc1,Z(0),&inc1);
    prscal (&k,&alpha,Z(0),&inc1);
    
    if (2 <= nZColumns)
    {
      finteger km1 = k-1;
      finteger ldHp1 = ldH+1;
      Complex help;
      /* z(1:k,2) = 0.5*(phiH*(H(1:k,1:k)*z(1:k,1))) + z(1:k,1) */
      pcopy (&k,Z(0),&inc1,Z(1),&inc1);
      
      ComplexSetReal(help,beta/2.0);
      if (k > 4)
	pgbmv ("N",&k,&k,&inc1,&km1,&help,H-km1,&ldHp1,Z(0),&inc1,
	      &zero,temp,&inc1);
      else
	pgemv ("No transpose",&k,&k,&help,H,&ldH,Z(0),&inc1,&zero,temp,&inc1);
      
      pgemv ("No transpose",&k,&k,&one,PhiH,&ldPhiH,temp,&inc1,
	    &one,Z(1),&inc1);
      
      if (3 <= nZColumns)
      {
	/* z(1:k,3)=2/3*(phiH*(H(1:k,1:k)*z(1:k,2))+z(1:k,2))+1/3*z(1:k,1) */
	pcopy (&k,Z(1),&inc1,Z(2),&inc1);
	
	ComplexSetReal(help,beta);
	
	if (k > 4)
	  pgbmv ("N",&k,&k,&inc1,&km1,&help,H-km1,&ldHp1,Z(1),&inc1,
		&zero,temp,&inc1);
	else
	  pgemv("No transpose",&k,&k,&help,H,&ldH,Z(1),&inc1,&zero,temp,&inc1);
	
	pgemv ("No transpose",&k,&k,&one,PhiH,&ldPhiH,temp,&inc1,
	      &one,Z(2),&inc1);
	
	prscal (&k,&two_three,Z(2),&inc1);
	
	praxpy (&k,&one_three,Z(0),&inc1,Z(2),&inc1);
      }
    }
  }
#undef Z
}

size_t PREC(SizePhie1HsDiagWork) (finteger roco,finteger fac_t)
     /* returns the workspace size needed by phim */
{
  finteger ld=roco;
  finteger lwork=ld*max(4,ld);

  size_t WorkSize = 2*GetWorkSpaceSize(sizeof(Complex)*roco*ld);

  WorkSize += GetWorkSpaceSize(sizeof(Complex)*roco*fac_t);
  WorkSize += GetWorkSpaceSize(sizeof(Complex)*roco);
  WorkSize += GetWorkSpaceSize(sizeof(Real)*roco);
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  WorkSize += 2*GetWorkSpaceSize(sizeof(Complex)*roco);
  WorkSize += 2*GetWorkSpaceSize(sizeof(finteger)*roco);
#else
  WorkSize += GetWorkSpaceSize(sizeof(Complex)*roco);
  WorkSize += GetWorkSpaceSize(sizeof(finteger)*2*roco);
  WorkSize += GetWorkSpaceSize(sizeof(finteger)*roco);
#endif
  WorkSize += GetWorkSpaceSize(sizeof(Complex)*lwork);
  return WorkSize;
}

int PREC(Exp4Phie1HsDiag) (WorkSpace Work,Complex *Z_,finteger LdZ,
      finteger fac_t,Complex *A,finteger LdA,finteger roco,Real fac,Real Alpha)

     /*
       Calculates E=(exp(fac*A)-Id)/(fac*A) by diagonalization.
       It is called by Exp4PhimHs.

       Parameters:

       mem       Is a pointer to the workspace.
                 (roco*roco+roco*max(4,roco)+2*roco)*sizeof(Complex)+
                 2*roco*sizeof(finteger) bytes are required.

       e         Pointer to the result matrix E.

       lde       Leading dimension of E.

       a_org     Pointer to the input matrix A. unchanged on exit.

       lda       Leading dimension of A.

       roco      Dimension of A and E.

       fac       Real factor.
       */

{
  Complex *TmpZ_;
  Complex *WorkA;
  Complex *Vr;
  Complex *work;
  Complex *Tau;
  Real *Scale;
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  Complex *EvRe,*EvIm;
#else
  Complex *Ev;
  Real *rwork;
#endif
  finteger ld=roco;
  finteger *perm;
  finteger lwork;
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  finteger *iwork;
#endif
  finteger info;
  finteger i;

#define  VR(i)     (Vr+(i)*ld)
#define  TmpZ(i,j) TmpZ_[(i)+(j)*ld]

  DistributeWorkspace(WorkA,sizeof(Complex)*roco*ld,Work);
  DistributeWorkspace(Vr,sizeof(Complex)*roco*ld,Work);
  DistributeWorkspace(TmpZ_,sizeof(Complex)*roco*fac_t,Work);
  DistributeWorkspace(Tau,sizeof(Complex)*roco,Work);
  DistributeWorkspace(Scale,sizeof(Real)*roco,Work);
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  DistributeWorkspace(EvRe,sizeof(Complex)*roco,Work);
  DistributeWorkspace(EvIm,sizeof(Complex)*roco,Work);
  DistributeWorkspace(iwork,sizeof(finteger)*roco,Work);
  DistributeWorkspace(perm,sizeof(finteger)*roco,Work);
#else
  DistributeWorkspace(Ev,sizeof(Complex)*roco,Work);
  DistributeWorkspace(rwork,2*sizeof(Real)*roco,Work);
  DistributeWorkspace(perm,sizeof(finteger)*roco,Work);
#endif
  lwork = GetRemainingWorkSpaceSize(Work)/sizeof(Complex);
  DistributeWorkspace(work,sizeof(Complex)*lwork,Work);

  /* copying A into WorkA Pade approximation */
  placpy ("Full",&roco,&roco,A,&LdA,WorkA,&ld);

  /* getting the eigenvalues and eigenvectors of A, eigenvectors are in Vr */
  /* this is adapted from dgeev.f to use the upper Hessenberg structure */
  {
    finteger ilo,ihi;
    finteger nout;
    flogical select;

#ifdef USE_HESSENBERG_FORM
    /* balancing A */
    pgebal ("S",&roco,WorkA,&ld,&ilo,&ihi,Scale,&info);
    lapack_err(info,"Exp4PhimHsDiag","gebal");

    /* do QR itteration */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    phseqr("S","I",&roco,&ilo,&ihi,WorkA,&ld,EvRe,EvIm,Vr,&ld,
	  work,&lwork,&info);
#else
    phseqr("S","I",&roco,&ilo,&ihi,WorkA,&ld,Ev,Vr,&ld,
	  work,&lwork,&info);
#endif
    lapack_err(info,"Exp4PhimHsDiag","hseqr");
#else /* USE_HESSENBERG_FORM */
    /* balancing A */
    pgebal ("B",&roco,WorkA,&ld,&ilo,&ihi,Scale,&info);
    lapack_err(info,"Exp4PhimHsDiag","gebal");
    
    /* reduce to upper hessenberg */
    pgehrd (&roco,&ilo,&ihi,WorkA,&ld,Tau,work,&lwork,&info);
    lapack_err(info,"Exp4PhimHsDiag","gehrd");
    
    placpy ("L",&roco,&roco,WorkA,&ld,Vr,&ld);
    
    porghr (&roco,&ilo,&ihi,Vr,&ld,Tau,work,&lwork,&info);
    lapack_err(info,"Exp4PhimHsDiag","orghr");
    
    /* do QR itteration */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    phseqr("S","V",&roco,&ilo,&ihi,WorkA,&ld,EvRe,EvIm,Vr,&ld,
	  work,&lwork,&info);
#else
    phseqr("S","V",&roco,&ilo,&ihi,WorkA,&ld,Ev,Vr,&ld,
	  work,&lwork,&info);
#endif
    lapack_err(info,"Exp4PhimHsDiag","hseqr");
#endif /* USE_HESSENBERG_FORM */

    /* get the eigenvectors */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    ptrevc ("R","B",&select,&roco,WorkA,&ld,NULL,&ld,Vr,&ld,&roco,&nout,
	  work,&info);
#else
    ptrevc ("R","B",&select,&roco,WorkA,&ld,NULL,&ld,Vr,&ld,&roco,&nout,
	  work,rwork,&info);
#endif
    lapack_err(info,"Exp4PhimHsDiag","trevc");
    
#ifdef USE_HESSENBERG_FORM
    /* backtransform the right eigenvectors (balancing) */
    pgebak ("S","R",&roco,&ilo,&ihi,Scale,&roco,Vr,&ld,&info);
    lapack_err(info,"Exp4PhimHsDiag","gebak");
#else /* USE_HESSENBERG_FORM */
    /* backtransform the right eigenvectors (balancing) */
    pgebak ("B","R",&roco,&ilo,&ihi,Scale,&roco,Vr,&ld,&info);
    lapack_err(info,"Exp4PhimHsDiag","gebak");
#endif /* USE_HESSENBERG_FORM */

    /* Normalize eigenvectors and make largest component real */
    for (i=0;i<roco;i++) {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
      if (EvIm[i] == 0.0) {
	Real scl = 1.0/pnrm2(&roco,VR(i),&inc1);
	pscal (&roco,&scl,VR(i),&inc1);
      } else {
	int j;
	Real normi = pnrm2(&roco,VR(i),&inc1);
	Real normi1 = pnrm2(&roco,VR(i+1),&inc1);
	Real scl = 1.0/sqrt(normi*normi+normi1*normi1);
	pscal (&roco,&scl,VR(i),&inc1);
	pscal (&roco,&scl,VR(i+1),&inc1);
	for (j=0;j<roco;j++)
	{
	  Real help1 = *(VR(i)+j);
	  Real help2 = *(VR(i+1)+j);
	  work[j] = help2*help2+help1*help1;
	}
	j = ipamax(&roco,work,&inc1)-1;
	{
	  Real cs,sn,r;
	  plartg(VR(i)+j,VR(i+1)+j,&cs,&sn,&r);
	  prot(&roco,VR(i),&inc1,VR(i+1),&inc1,&cs,&sn);
	  (*(VR(i+1)+j)) = 0.0;
	}
	i++;
      }
#else
      Real scl = 1.0/pnrm2(&roco,VR(i),&inc1);
      prscal (&roco,&scl,VR(i),&inc1);
#endif
    }
  }

#if (DEBUG >= LOCAL_DEBUG_OFFSET+1)
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  {
    Real help = EvRe[0];
    for (i=1;i<roco;i++)
      if (help<EvRe[i])
	help = EvRe[i];
    Exp4Message2((DEBUG_FILE,"The largest realpart of the eigenvalues is: %e\n"
	  ,help*fac));
  }
#else
  {
    Real help = Re(Ev[0]);
    for (i=1;i<roco;i++)
      if (help<Re(Ev[i]))
	help = Re(Ev[i]);
    Exp4Message2((DEBUG_FILE,"The largest realpart of the eigenvalues is: %e\n"
	  ,help*fac));
  }
#endif
#endif

  /*
   * copying the eigenvalue matrix,
   * E will be multiplied by phi(z), z the eigenvalues
   */
  placpy ("Full",&roco,&roco,Vr,&ld,WorkA,&ld);

  {
    Real rcond;
#ifdef USE_QR_DECOMP_IN_DIAG
    for (i=0;i<roco;perm[i++]=0);
    pgeqpf (&roco,&roco,Vr,&ld,perm,Tau,work,&info);
    lapack_err(info,"Exp4PhimHsDiag","geqpf");

#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    ptrcon ("Inf","Upper","No Unit",&roco,Vr,&ld,&rcond,work,
	  iwork,&info);
#else
    ptrcon ("Inf","Upper","No Unit",&roco,Vr,&ld,&rcond,work,
	  rwork,&info);
#endif
#else /* case to use LU - decomposition */
    /* calculating the inf-norm of B, pgecon needs that value */
    Real norm = plange ("Inf",&roco,&roco,Vr,&ld,(Real*)work);

    /* computes the LU factors of B for getting the condition number of A
       and for calculating .../B  */
    pgetrf (&roco,&roco,Vr,&ld,&perm[0],&info);
    lapack_err(info,"Exp4PhimHs","getrf");

    /* gets the reciprocal condition number of B */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    pgecon ("Inf",&roco,Vr,&ld,&norm,&rcond,work,iwork,&info);
#else
    pgecon ("Inf",&roco,Vr,&ld,&norm,&rcond,work,
	  (Real*)(work+2*roco),&info);
#endif
    lapack_err(info,"Exp4PhimHs","gecon");
#endif

    Exp4Message1((DEBUG_FILE,"The condition number of the matrix of "
	  "eigenvalues: %e\n",1./rcond));

    if (1.0 < MAX_COND * rcond)
    {
      Exp4Message1((DEBUG_FILE,"Used diagonalization for calculating "
	    "Exp4PhimHs(Z)=(exp(Z)-1)/Z\n"));
      
      {
#ifdef USE_QR_DECOMP_IN_DIAG
	flogical flag = FFALSE;

	pcopy (&roco,&zero,&inc0,&TmpZ(0,0),&inc1);
	ComplexSetReal(TmpZ(0,0),1.0);

	/* FIXME are all elementary reflectors used when multiplying with e1?*/
	pormqr ("Left",PORMQR_TRANSPOSE,&roco,&inc1,&roco,Vr,&ld,Tau,
	      &TmpZ(0,0),&ld,work,&lwork,&info);
	lapack_err(info,"Exp4PhimHsDiag","ormqr");

	ptrsm ("Left","Upper","No transpose","No Unit",&roco,&inc1,
	      &one,Vr,&ld,&TmpZ(0,0),&ld);
	
	plapmt (&flag,&inc1,&roco,&TmpZ(0,0),&inc1,perm);
#else /* case to use LU - decomposition */
	pcopy (&roco,&zero,&inc0,&TmpZ(0,0),&inc1);
	ComplexSetReal(TmpZ(0,0),1.0);
	pgetrs ("N",&roco,&inc1,Vr,&ld,perm,&TmpZ(0,0),&ld,&info);
	lapack_err(info,"Exp4PhimHsDiag","getrs");
#endif
      }

      for (i=0;i<roco;i++)
      {
	/* calculate A=B*diag((exp(eigenvalues)-1)/eigenvalues) */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	if (EvIm[i] == 0.0)
	{
	  /* if the eigenvalue i is real then calculate it with
	   * real arithmetic */
	  int j;
	  for (j=fac_t;0<j;j--)
	    TmpZ(i,j-1) = TmpZ(i,0) * PREC(phi) (EvRe[i]*fac*j);
	}
	else
	{
	  /* if the eigenvalue i is complex calculate it with
	   * complex arithmetic */
	  int j;

	  for (j=fac_t;0<j;j--)
	  {
	    Real tmp;
	    fdoublecomplex chelp;
	    chelp.r =  EvRe[i]*fac*j;
	    chelp.i = -EvIm[i]*fac*j;
	    chelp   = Zphi (chelp);

	    tmp           = TmpZ(i,0) * chelp.i + TmpZ(i+1,0) * chelp.r;
	    TmpZ(i,j-1)   = TmpZ(i,0) * chelp.r - TmpZ(i+1,0) * chelp.i;
	    TmpZ(i+1,j-1) = tmp;
	  }
	  i++;     /* two eigenvalues are done in the imaginary case */
	}
#else
	int j;
	for (j=fac_t;0<j;j--)
	{
	  Complex help;
	  rmul(help,fac*j,Ev[i]);
	  help = PREC(phi) (help);
	  mul(TmpZ(i,j-1),TmpZ(i,0),help);
	}
#endif
      } /* for all eigenvalues */

      {
	Complex help;
	ComplexSetReal(help,Alpha);
	pgemm ("N","N",&roco,&fac_t,&roco,&help,WorkA,&ld,TmpZ_,&ld,
	      &zero,Z_,&LdZ);
      }
      return 1;
    }
  }
  return 0;

#undef VR
}
