/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include "ftypes.h"
#include "errors.h"
#include "debug.h"

#include "work.h"

#ifdef MEX
#include "mex.h"
#define malloc(size) mxMalloc(size)
#define free(ptr)    mxFree(ptr)
#endif

struct MemoryPackage
{
  struct MemoryPackage *Last;
  void *Data;
};

static struct MemoryPackage *Memory = NULL;
static size_t nBytes = 0l;

WorkSpace Exp4GetWorkspace(size_t WorkSize)
{
  WorkSpace RetVal;
  char *Data,*AlignData;

  /* calculate the number of bytes which are requested */
  RetVal.Length = GetWorkSpaceSize(WorkSize);

  Data = Exp4GetWorkPointer(RetVal.Length+ALIGN);

  /* does aligning to ALIGN boundaries */
  AlignData = (((char*)NULL) +
	(((((char*)Data)-(char*)NULL)/(ALIGN/sizeof(char)))*
	      (ALIGN/sizeof(char))));
  if (Data != AlignData)
    RetVal.Data = (AlignData + (ALIGN/sizeof(char)));
  else
    RetVal.Data = AlignData;

#if (WORK_DEBUG >= 2)
  fprintf(stdout,"WORK_DEBUG: Alligning: old: %p new: %p\n",Data,RetVal.Data);
#endif
  return RetVal;
}

void *Exp4GetWorkPointer(size_t Bytes)
{
  struct MemoryPackage *Actual;

  /* get workspace for the linked list */
  Actual = (struct MemoryPackage *)malloc(sizeof(struct MemoryPackage));
#if (WORK_DEBUG >= 2)
  fprintf(stdout,"WORK_DEBUG: allocating memory for storing allocation "
	"data. Addr = %p\n",Actual);
#endif

  /* statistics */
  if (Memory == NULL) nBytes = sizeof(struct MemoryPackage);
  else nBytes += sizeof(struct MemoryPackage);

  /* if we can't get workspace, do an error exit */
  if (Actual == NULL) ErrorReturn("Can't allocate memory!",EXP4_ERR_NO_MEM);

  Actual->Data = malloc(Bytes);
#if (WORK_DEBUG >= 2)
  fprintf(stdout,"WORK_DEBUG: allocating requested memory: %d bytes = %g "
	"K = %g Meg. Addr = %p\n",Bytes,(double)(Bytes/1024.0),
	(double)(Bytes/1024.0/1024.0),Actual->Data);
#endif
  nBytes += Bytes;

  if (Actual->Data == NULL)
  {
    free(Actual);
    ErrorReturn("Can't allocate memory!",EXP4_ERR_NO_MEM);
  }

  if (Memory == NULL)
  {
#if (WORK_DEBUG >= 2)
    fprintf(stdout,"WORK_DEBUG: writing the first MemoryPackage, setting "
	  "Last to NULL, Memory to %p\n",Actual->Data);
#endif
    Actual->Last = NULL;
  }
  else
  {
#if (WORK_DEBUG >= 2)
    fprintf(stdout,"WORK_DEBUG: writing the first MemoryPackage, setting "
	  "Last to %p, Memory to %p\n",Memory,Actual->Data);
#endif
    Actual->Last = Memory;
  }
  Memory = Actual;

#if (WORK_DEBUG >= 2)
  fprintf(stdout,"WORK_DEBUG: leaving Exp4GetWorkspace, Memory = %p, return "
	"value = %p\n",Memory,Actual->Data);
#endif

  return Actual->Data;
}

void Exp4FreeAllWorkspace()
{
  while (Memory != NULL)
  {
    struct MemoryPackage *LastPackage = Memory->Last;
#if (WORK_DEBUG >= 2)
    fprintf(stdout,"WORK_DEBUG: freeing Pair Data = %p, Memory = %p\n",
	  Memory->Data,Memory);
#endif
    free (Memory->Data);
    free (Memory);

    Memory = LastPackage;
  }
#if (WORK_DEBUG >= 2)
  fprintf(stdout,"WORK_DEBUG: leaving Exp4FreeAllWorkspace, Memory = %p\n",
	Memory);
#endif
}

double AmountOfMemory()
{
  if (nBytes < 1024) return (double)nBytes;
  if (nBytes < (1024*1024)) return (double)(nBytes/1024.0);
  return (double)(nBytes/(1024.0*1024.0));
}

char AmountOfMemoryUnit()
{
  if (nBytes < 1024) return ' ';
  if (nBytes < (1024*1024)) return 'K';
  return 'M';
}
