/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

/* header file for work.c */

#ifndef WORK_H
#define WORK_H

#include "errors.h"

/* #define WORK_DEBUG 5 */
#define WORK_DEBUG 1

#ifndef WORK_DEBUG
#define WORK_DEBUG 0
#endif

#ifdef max
#undef max
#endif
#define max(a,b) (((a)<(b))?(b):(a))

/*
 * The alignment size has to be a power of two.
 */
#define ALIGN 32

#ifndef ALIGN
#define ALIGN (max(sizeof(finteger),sizeof(fdoublecomplex)))
#endif

#define ALIGN_MASK ((size_t)(ALIGN-1))

#define GetRemainingWorkSpaceSize(Work)					      \
((~ALIGN_MASK) & ((Work).Length*sizeof(char)))

#define GetWorkSpaceSize(Size) ((size_t)((~ALIGN_MASK) & (ALIGN_MASK+(Size))))

typedef struct { size_t Length; char *Data; } WorkSpace;

#if (WORK_DEBUG >= 1)

#define DistributeWorkspace(Ptr,Size,Work)				      \
{									      \
    size_t __Size = GetWorkSpaceSize(Size);				      \
    (Ptr) = (void*)(Work).Data;						      \
    (Work).Data += __Size/sizeof(char);					      \
    if ((Work).Length < __Size)						      \
    {									      \
      ErrorReturn("Error in workspace distribution. Length of Workspace"      \
	    " is less then zero!",EXP4_ERR_IN_WORKSPACE_DISTRIBUTION);	      \
    }									      \
    (Work).Length -= __Size;						      \
}

#define ShareWorkspace(Ptr,Size,Work)					      \
{									      \
    size_t __Size = GetWorkSpaceSize(Size);				      \
    (Ptr) = (void*)(Work).Data;						      \
    if ((Work).Length < __Size)						      \
    {									      \
      ErrorReturn("Error in workspace distribution. Length of Workspace"      \
	    " is less then zero!",EXP4_ERR_IN_WORKSPACE_DISTRIBUTION);	      \
    }									      \
}

#define GiveBackWorkspace(Ptr,Size,Work)				      \
{									      \
    size_t __Size = GetWorkSpaceSize(Size);				      \
    if (((size_t)((Work).Data-((char*)(Ptr)))) != (__Size/sizeof(char)))      \
    {									      \
      ErrorReturn("Error in workspace distribution. Freeing invalid "	      \
	    "pointer!", EXP4_ERR_IN_WORKSPACE_DISTRIBUTION);		      \
    }									      \
    (Work).Data -= __Size/sizeof(char);					      \
    (Work).Length += __Size;						      \
}

#define ReDistributeWorkspace(Ptr,OldSize,NewSize,Work)			      \
{									      \
    size_t __OldSize = GetWorkSpaceSize(OldSize);			      \
    size_t __NewSize = GetWorkSpaceSize(NewSize);			      \
    if (((size_t)((Work).Data-((char*)(Ptr)))) != (__OldSize/sizeof(char)))   \
    {									      \
      ErrorReturn("Error in workspace distribution. Freeing invalid "	      \
	    "pointer!", EXP4_ERR_IN_WORKSPACE_DISTRIBUTION);		      \
    }									      \
    (Work).Data += (__NewSize-__OldSize)/sizeof(char);			      \
    if (((Work).Length + __OldSize) < __NewSize)			      \
    {									      \
      ErrorReturn("Error in workspace distribution. Length of Workspace"      \
	    " is less then zero!",EXP4_ERR_IN_WORKSPACE_DISTRIBUTION);	      \
    }									      \
    (Work).Length -= __NewSize - __OldSize;				      \
}

#else

#define DistributeWorkspace(Ptr,Size,Work)				      \
{									      \
    size_t __Size = GetWorkSpaceSize(Size);				      \
    (Ptr) = (void*)(Work).Data;						      \
    (Work).Data += __Size/sizeof(char);					      \
    (Work).Length -= __Size;						      \
}

#define ShareWorkspace(Ptr,Size,Work)					      \
{									      \
    (Ptr) = (void*)(Work).Data;						      \
}

#define GiveBackWorkspace(Ptr,Size,Work)				      \
{									      \
    size_t __Size = GetWorkSpaceSize(Size);				      \
    (Work).Data -= __Size/sizeof(char);					      \
    (Work).Length += __Size;						      \
}

#define ReDistributeWorkspace(Ptr,OldSize,NewSize,Work)			      \
{									      \
    size_t __OldSize = GetWorkSpaceSize(OldSize);			      \
    size_t __NewSize = GetWorkSpaceSize(NewSize);			      \
    (Work).Data += (__NewSize - __OldSize)/sizeof(char);		      \
    (Work).Length -= __NewSize - __OldSize;				      \
}

#endif

#ifdef  __cplusplus
extern "C" {
#endif

extern WorkSpace Exp4GetWorkspace(size_t Bytes);

extern void *Exp4GetWorkPointer(size_t Bytes);

extern void Exp4FreeAllWorkspace();

extern double AmountOfMemory();

extern char AmountOfMemoryUnit();

#ifdef  __cplusplus
}
#endif
#endif
