/*************************************************************************/
/* Copyright (C) 1996 by Mathias Froehlich.                              */
/*                                                                       */
/* All rights reserved.                                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYRIGHT" in the this directory.                                    */
/*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include "undersc.h"

#if (defined(NO_F77_UNDERSCORING) && !defined(USE_F2C))
#define F77CALL(CallName)  CallName
#endif
#if (defined(F77_UNDERSCORING) || defined(USE_F2C))
#define F77CALL(CallName)  CallName##_
#endif

#define REALTYPE " fsingle "
#define DOUBLETYPE " fdouble "
#define INTEGERTYPE " finteger "
#define LOGICALTYPE " flogical "
#define COMPLEXTYPE " fsinglecomplex "
#define DOUBLECOMPLEXTYPE " fdoublecomplex "

#define DATA_LENGTH 100

extern void F77CALL (finteger) (long *);
extern void F77CALL (flogical) (long *);
extern void F77CALL (freal) (long *);
extern void F77CALL (fdoublereal) (long *);

typedef struct { double r,i; } fdoublecomplex;

static jmp_buf env;
static fdoublecomplex ret;

extern void F77CALL (complexret) (void *);
void F77CALL (structcall) (void *,void *);


int int_type(FILE *file,long *data);
int logic_type(FILE *file,long *data);
int float_type(FILE *file,long *data);
int ret_complex_by_pointer(long *data);
void fill_data(long *data);


int main()
{
  long data[DATA_LENGTH];
  FILE *file;

  if (NULL == (file=fopen("ftypes.h","w")))
  {
    fprintf(stderr,"problems opening file for output ...\nAbort ...\n");
    return EXIT_FAILURE;
  }

  fprintf(stdout,"\nftypes V 1.0 (c) Mathias Froehlich 1996\n\nThis program "
	  "generates automaticaly the header file ftypes.h.\nThe file "
	  "ftypes.h containes the definitions for the c types\nwhich are "
	  "aquivalent to those, used by the Fortran BLAS\nand LAPACK "
	  "routines. We can't take f2c.h on all systems ...\n\n");

  fprintf(file,
	  "/*                    Header file ftypes.h                   */\n"
	  "/* generated by ftypes V 1.0 (c) Mathias Froehlich 1996      */\n"
	  "/* This types are needed for calling Fortran BLAS and LAPACK */\n"
	  "/* routines.                                                 */\n"
	  "\n\n#ifndef FTYPES_H_\n#define FTYPES_H_\n\n"
	  "#define FTRUE (1)\n#define FFALSE (0)\n\n");

#if (defined(NO_F77_UNDERSCORING) && !defined(USE_F2C))
  fprintf(file,"#define F77CALL(name)  name\n\n");
#endif
#if (defined(F77_UNDERSCORING) || defined(USE_F2C))
  fprintf(file,"#define F77CALL(name)  F77CAT(name,_,)\n\n");
#endif
  fprintf(file,"#define F77CAT(a,b,c)  a##b##c\n\n");

#ifdef USE_F2C
  fprintf(file,"#include \"f2c.h\"\n\n");
#endif


  fill_data(data);
  F77CALL (finteger) (data);
  fprintf(file,"typedef ");
#ifdef USE_F2C
  fprintf(file," integer ");
#else
  if (int_type(file,data) < 0)
  {
    fclose(file);
    return EXIT_FAILURE;
  }
#endif
  fprintf(file,INTEGERTYPE";\n");


  fill_data(data);
  F77CALL (flogical) (data);
  fprintf(file,"typedef ");
#ifdef USE_F2C
  fprintf(file," logical ");
#else
  if (logic_type(file,data) < 0)
  {
    fclose(file);
    return EXIT_FAILURE;
  }
#endif
  fprintf(file,LOGICALTYPE";\n");


  fill_data(data);
  F77CALL (freal) (data);
  fprintf(file,"typedef ");
#ifdef USE_F2C
  fprintf(file," float ");
#else
  if (float_type(file,data) < 0)
  {
    fclose(file);
    return EXIT_FAILURE;
  }
#endif
  fprintf(file,REALTYPE";\n");


  fill_data(data);
  F77CALL (fdoublereal) (data);
  fprintf(file,"typedef ");
#ifdef USE_F2C
  fprintf(file," doublereal ");
#else
  if (float_type(file,data) < 0)
  {
    fclose(file);
    return EXIT_FAILURE;
  }
#endif
  fprintf(file,DOUBLETYPE";\n");

  fprintf(file,
	  "\ntypedef struct { "REALTYPE" r, i; } "COMPLEXTYPE";\n"
	  "typedef struct { "DOUBLETYPE" r, i; } "DOUBLECOMPLEXTYPE";\n\n");

#ifdef USE_F2C
  fprintf(file,"#define USE_F2C\n\n");
#else
  if (ret_complex_by_pointer(data))
    fprintf(file,"#define RETURN_COMPLEX_WITH_POINTER\n\n");
#endif

  fprintf(file,"#endif /* FTYPES_H_ */\n");

  fclose(file);
  return EXIT_SUCCESS;
}

int int_type(FILE *file,long *data)
{
  short *ShortPtr,ShortRef=~0;
  int *IntPtr,IntRef=~0;
  long *LongPtr,LongRef=~0l;

  ShortPtr=(short *)data;
  if ((ShortPtr[0] == 0) && (ShortPtr[1] == 1) && (ShortPtr[2] == -1)
	&& (ShortPtr[3] == ShortRef) && (ShortPtr[4] == ShortRef) &&
	(ShortPtr[5] == ShortRef) && (ShortPtr[6] == 0) && (ShortPtr[7] == 1)
	&& (ShortPtr[8] == -1))
  {
    fprintf(file," short ");
    return 1;
  }

  IntPtr=(int *)data;
  if ((IntPtr[0] == 0) && (IntPtr[1] == 1) && (IntPtr[2] == -1)
	&& (IntPtr[3] == IntRef) && (IntPtr[4] == IntRef) &&
	(IntPtr[5] == IntRef) && (IntPtr[6] == 0) && (IntPtr[7] == 1) &&
	(IntPtr[8] == -1))
  {
    fprintf(file," int ");
    return 2;
  }

  LongPtr=(long *)data;
  if ((LongPtr[0] == 0) && (LongPtr[1] == 1) && (LongPtr[2] == -1)
	&& (LongPtr[3] == LongRef) && (LongPtr[4] == LongRef) &&
	(LongPtr[5] == LongRef)	&& (LongPtr[6] == 0) && (LongPtr[7] == 1) &&
	(LongPtr[8] == -1))
  {
    fprintf(file," long ");
    return 3;
  }

  return -1;
}

int logic_type(FILE *file,long *data)
{
  short *ShortPtr,ShortRef=~0;
  int *IntPtr,IntRef=~0;
  long *LongPtr,LongRef=~0l;

  ShortPtr=(short *)data;
  if ((ShortPtr[0] == 1) && (ShortPtr[1] == 0) && (ShortPtr[2] == 1)
	&& (ShortPtr[3] == ShortRef) && (ShortPtr[4] == ShortRef) &&
	(ShortPtr[5] == ShortRef) && (ShortPtr[6] == 1) && (ShortPtr[7] == 0)
	&& (ShortPtr[8] == 1))
  {
    fprintf(file," short ");
    return 1;
  }

  IntPtr=(int *)data;
  if ((IntPtr[0] == 1) && (IntPtr[1] == 0) && (IntPtr[2] == 1)
	&& (IntPtr[3] == IntRef) && (IntPtr[4] == IntRef) &&
	(IntPtr[5] == IntRef) && (IntPtr[6] == 1) && (IntPtr[7] == 0) &&
	(IntPtr[8] == 1))
  {
    fprintf(file," int ");
    return 2;
  }

  LongPtr=(long *)data;
  if ((LongPtr[0] == 1) && (LongPtr[1] == 0) && (LongPtr[2] == 1)
	&& (LongPtr[3] == LongRef) && (LongPtr[4] == LongRef) &&
	(LongPtr[5] == LongRef)	&& (LongPtr[6] == 1) && (LongPtr[7] == 0) &&
	(LongPtr[8] == 1))
  {
    fprintf(file," long ");
    return 3;
  }

  return -1;
}

int float_type(FILE *file,long *data)
{
  float *FloatPtr;
  double *DoublePtr;

  DoublePtr=(double *)data;
  if ((DoublePtr[0] == 0.0) && (DoublePtr[1] == 1.0) && (DoublePtr[2] == -1.0)
	&& (DoublePtr[6] == 0.0) && (DoublePtr[7] == 1.0) && 
	(DoublePtr[8] == -1.0))
  {
    fprintf(file," double ");
    return 1;
  }

  FloatPtr=(float *)data;
  if ((FloatPtr[0] == 0.0) && (FloatPtr[1] == 1.0) && (FloatPtr[2] == -1.0)
	&& (FloatPtr[6] == 0.0) && (FloatPtr[7] == 1.0) &&
	(FloatPtr[8] == -1.0))
  {
    fprintf(file," float ");
    return 2;
  }
  return -1;
}

int ret_complex_by_pointer(long *data)
{
  int type;
  int ret_val;
  FILE *file = tmpfile();

  fill_data(data);
  F77CALL (fdoublereal) (data);
  type = float_type(file,data);
  fclose(file);
  
  if ((ret_val = setjmp(env)) == 0)
    F77CALL (complexret) (&ret);
  else
  {
    if (ret_val == 1)
      return 1;
    else
      return 0;
  }

  /* never reached, this line supresses warnings */
  return -1;
}

void F77CALL (structcall) (void *ptr1,void *ptr2)
{
  if (((void *)&ret) == ptr2)
    longjmp(env,1);
  else
    longjmp(env,2);
}

void fill_data(long *data)
{
  int i;
  for (i=0;i<DATA_LENGTH;data[i++]=~0l);
}
